/*-----------------------------------------------------------------------

File  : ccl_clausesets.h

Author: Stephan Schulz

Contents
 
  Definitions dealing with collections of clauses

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Sat Jul  5 02:28:25 MET DST 1997
    New

-----------------------------------------------------------------------*/

#ifndef CCL_CLAUSESETS

#define CCL_CLAUSESETS

#include <ccl_inferencedoc.h>
#include <ccl_pdtrees.h>
#include <clb_objtrees.h>


/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/

/* Clause sets are doubly linked lists of clauses with indices for the
   various potential evaluations. */

typedef struct clausesetcell
{
   long      members; /* How many clauses are there? */
   long      literals; /* And how many literals? */
   Clause_p  anchor;  /* The clauses */
   SysDate   date;    /* Age of the clause set, used for optimizing
		 	 rewriting. The special date SysCreationDate()
			 is used to indicate ignoring of dates when
			 checking for irreducability. */
   PDTree_p  demod_index; /* If used for demodulators */
   PDArray_p eval_indices;
}ClauseSetCell, *ClauseSet_p;


/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/


#define ClauseSetCellAlloc()    (ClauseSetCell*)SizeMalloc(sizeof(ClauseSetCell))
#define ClauseSetCellFree(junk) SizeFree(junk, sizeof(ClauseSetCell))

#define     ClauseSetStorage(set) (4*(set)->members+(set)->literals)
ClauseSet_p ClauseSetAlloc();
void        ClauseSetFree(ClauseSet_p junk);
void        ClauseSetInsert(ClauseSet_p set, Clause_p new);
void        ClauseSetIndexedInsert(ClauseSet_p set, Clause_p new);
void        ClauseSetExtractEntry(Clause_p clause);
void        ClauseSetIndexedExtractEntry(Clause_p clause);
#define     ClauseSetEmpty(set)\
            ((set)->anchor->succ == (set)->anchor)
Clause_p    ClauseSetExtractFirst(ClauseSet_p set);
void        ClauseSetDeleteEntry(Clause_p clause);
void        ClauseSetIndexedDeleteEntry(Clause_p clause);
Clause_p    ClauseSetFindBest(ClauseSet_p set, int index);
void        ClauseSetPrint(FILE* out, ClauseSet_p set, bool
			   fullterms);
void        ClauseSetSetProp(ClauseSet_p set, ClauseProperties prop);
void        ClauseSetDelProp(ClauseSet_p set, ClauseProperties prop);

long        ClauseSetMarkCopies(ClauseSet_p set);
long        ClauseSetDeleteMarkedEntries(ClauseSet_p set);
long        ClauseSetDeleteCopies(ClauseSet_p set);
long        ClauseSetDeleteNonUnits(ClauseSet_p set);

long        ClauseSetGetTermNodes(ClauseSet_p set);
long        ClauseSetParseList(Scanner_p in, ClauseSet_p set, TB_p bank);
void        ClauseSetMarkMaximalTerms(OCB_p ocb, ClauseSet_p set);
SysDate     ClauseSetListGetMaxDate(ClauseSet_p *demodulators, int
				    limit);
void        ClauseKillChildren(Clause_p clause);
int         ClauseRemoveSuperfluousLiterals(Clause_p clause);
Clause_p    ClauseSetFind(ClauseSet_p set, Clause_p clause);
Clause_p    ClauseSetFindById(ClauseSet_p set, long ident);
void        ClauseSetRemoveEvaluations(ClauseSet_p set);
long        ClauseSetFilterTautologies(ClauseSet_p set);
bool        ClauseSetVerifyDemod(ClauseSet_p demods, ClausePos_p pos);

bool        PDTreeVerifyIndex(PDTree_p root, ClauseSet_p demods);

void        ClauseSetDocumentProcessing(FILE* out, long level,
					ClauseSet_p set, char*
					operation);

void        EqAxiomsPrint(FILE* out, Sig_p sig, bool single_subst);


void        ClauseSetAddSymbolDistribution(ClauseSet_p set, long
				       *dist_array);

#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





