/*-----------------------------------------------------------------------

File  : cto_orderings.c

Author: Stephan Schulz

Contents
 
  Interface to the ordering module

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Mon May  4 23:24:41 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "cto_orderings.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: TOGreater()
//
//   Test wether t1 is greater that t2 in the ordering described by
//   the ocb.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool TOGreater(OCB_p ocb, Term_p s, Term_p t, DerefType deref_s,
	       DerefType deref_t)
{
   bool res = false;
#ifdef NEVER_DEFINED
   bool res2 = false;
#endif

   assert(ocb);
   assert(s);
   assert(t);

#ifdef NEVER_DEFINED   
   printf("\n#################################################\n");
   printf("s (ONCE)   =");
   TermPrint(stdout, s, ocb->sig, DEREF_ONCE);
   printf("\ns (deref_s)=");
   TermPrint(stdout, s, ocb->sig, deref_s);
   printf("\ns (NEVER)  =");
   TermPrint(stdout, s, ocb->sig, DEREF_NEVER);
   
   printf("\nt (ONCE)   =");
   TermPrint(stdout, t, ocb->sig, DEREF_ONCE);
   printf("\nt (deref_t)=");
   TermPrint(stdout, t, ocb->sig, deref_t);
   printf("\nt (NEVER)  =");
   TermPrint(stdout, t, ocb->sig, DEREF_NEVER);
   printf("\n");
#endif

   switch(ocb->type)
   {
   case LPO:
	 res = LPOGreater(ocb, s, t, deref_s, deref_t);
#ifdef NEVER_DEFINED
	 if (LPOCompare(ocb, s, t, deref_s, deref_t) == to_greater)
	 {
	    res2 = true;
	 }
	 if (res) {
	    printf("Real LPOGreater(): true\n");
	 }
	 else {
	    printf("Real LPOGreater(): false\n");
	 }
	 if (res2) {
	    printf("Via LPOCompare() : true\n");
	 }
	 else {
	    printf("Via LPOCompare() : false\n");
	 }
	 assert(!XOR(res,res2));
#endif

	 break;
   case RPO:
	 assert(false && "RPO not yet implemented!");
	 break;
   case KBO:
	 res = KBOGreater(ocb, s, t, deref_s, deref_t); 
#ifdef NEVER_DEFINED
	 res2 = KBOCompare(ocb, s, t, deref_s, deref_t);
	 printf("KBOGreater(): %d\n", res);
	 printf("KBOCompare(): %s ==> %d\n", TOCompareSymbol[res2],
		(res2==to_greater));
	 res2 = (res2==to_greater);
	 assert(!XOR(res,res2));		 
#endif
	 break;
   case KBO2:
	 res = KBOGreaterNew(ocb, s, t, deref_s, deref_t); 
	 break;
   default:
	 assert(false);
	 break;
   }
#ifdef NEVER_DEFINED
   printf("TOGreater: %d\n", res);
   printf("#################################################\n");
#endif
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TOCompare()
//
//   Compare t1 and t2 in the ordering described by the ocb.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

CompareResult TOCompare(OCB_p ocb, Term_p s, Term_p t, DerefType deref_s,
	       DerefType deref_t)
{
   CompareResult res = to_uncomparable;

   assert(ocb);
   assert(s);
   assert(t);

   switch(ocb->type)
   {
   case LPO:
	 res = LPOCompare(ocb, s, t, deref_s, deref_t);
	 break;
   case RPO:
	 assert(false && "RPO not yet implemented!");
	 break;
   case KBO:
	 res = KBOCompare(ocb, s, t, deref_s, deref_t);
	 break;
   case KBO2:
	 res = KBOCompareNew(ocb, s, t, deref_s, deref_t);
	 break;
   default:
	 assert(false);
	 break;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TOSymbolComparisonParse()
//
//   Parse a declaration f>g (f=g, f<g) and insert it into ocb. Return
//   new OCB status pointer.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

PStackPointer TOSymbolComparisonParse(Scanner_p in, OCB_p ocb, Sig_p sig)
{
   FunCode       f1, f2; 
   int           line, column;
   CompareResult relation;
   DStr_p        source_name, errpos;
   PStackPointer ocb_state;
   StreamType    type;

   line = AktToken(in)->line;
   column = AktToken(in)->column;
   source_name = DStrGetRef(AktToken(in)->source);
   type = AktToken(in)->stream_type;

   f1 = SigParseKnownOperator(in, sig);

   CheckInpTok(in, LesserSign|GreaterSign|EqualSign);
   switch(AktToken(in)->tok)
   {
   case LesserSign:
	 relation = to_lesser;
	 break;
   case GreaterSign:
	 relation = to_greater;
	 break;
   case EqualSign:
	 relation = to_equal;
	 break;
   default:
	 assert(false);
	 relation = to_uncomparable;
	 break;
   }
   NextToken(in);
   
   f2 = SigParseKnownOperator(in, sig);
   
   ocb_state = OCBPrecedenceAddTuple(ocb, f1, f2, relation);
   
   if(!ocb_state)
   {
      errpos = DStrAlloc();
      
      DStrAppendStr(errpos, PosRep(type, source_name, line, column));
      DStrAppendStr(errpos, " Precedence incompatible with previous ordering!");
      Error(DStrView(errpos), SYNTAX_ERROR);
      DStrFree(errpos);
   }
   DStrReleaseRef(source_name);
   
   return ocb_state;
}


/*-----------------------------------------------------------------------
//
// Function: TOPrecedenceParse()
//
//   Parse a precedence (list of comparisons).
//
// Global Variables: -
//
// Side Effects    : Changes OCB, reads input
//
/----------------------------------------------------------------------*/

PStackPointer TOPrecedenceParse(Scanner_p in, OCB_p ocb, Sig_p sig)
{
   PStackPointer res;

   assert(ocb);
   assert(ocb->sig_size == sig->f_count);
   assert(ocb->precedence); 

   res = OCBPrecedenceGetState(ocb);
   while(!TestInpTok(in, Fullstop))
   {
      res = TOSymbolComparisonParse(in, ocb, sig);
   }
   AcceptInpTok(in, Fullstop);

   return res;
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


